import Component from '@glimmer/component';
import { tracked } from '@glimmer/tracking';
import { bisector } from "d3-array";
import { throttle } from '@ember/runloop';

export default class MultiScrubberComponent extends Component {
  get bisectDate() {
    return bisector(d => typeof d[this.dateKey] === "string" ? new Date(d[this.dateKey]) : d[this.dateKey]).left;
  }

  @tracked fadeState = "opacity-0";
  @tracked currentMousePos;
  @tracked cornerTextValues;
  @tracked line;
  @tracked movingTextContainer;
  @tracked textBox;
  @tracked animatedTextXPosition;
  @tracked textSizeElement;
  @tracked hoverTextValues;
  // ~~~~~~~~~~~~~~~~~~~~
  // CONFIGURATIONS
  // ~~~~~~~~~~~~~~~~~~~~
  get strokeWidth() {
    return this.args.width ? this.args.width : '1';
  }

  get lineColor() {
    return this.args.lineColor || 'gray';
  }

  get lineTextUnits() {
    return this.args.lineTextUnits || '';
  }

  get cornerTextUnits() {
    return this.args.cornerTextUnits || '';
  }

  get valTextKey() {
    return this.args.valTextKey || 'value';
  }

  get dateKey() {
    return this.args.dateKey || 'at';
  }

  get singleValueCornerText() {
    return this.args.singleValueCornerText || false;
  }

  get textBoxFillOpacity() {
    return this.args.textBoxFillOpacity || '1';
  }

  get textBoxFill() {
    return this.args.textBoxFill || 'white';
  }

  get textLineHeight() {
    return this.args.textLineHeight || 15
  }

  get textBoxPadding() {
    return this.args.textBoxPadding || 20;
  }

  get hideDate() {
    return this.args.hideDate || false;
  }

  get hideTime() {
    return this.args.hideTime || false;
  }

  get putValueInCorner() {
    return this.args.putValueInCorner || false;
  }

  get cornerTextOffset() {
    return this.args.cornerTextOffset || '150px';
  }

  // ~~~~~~~~~~~~~~~~~~~~
  // COMPUTED
  // ~~~~~~~~~~~~~~~~~~~~
  get primaryDataset() {
      return this.args.datasets ? this.args.datasets[0] : [];
  }
  
  get svg() {
    return this.line ? this.findSvg() : null;
  }

  get svgDimensions() {
    return this.svg ? this.svg.getBoundingClientRect() : {};  
  }

  get chartWidth() {
    return this.args.chartWidth || this.svgDimensions.width;
  }

  get lineHeight() {
    return this.svgDimensions.height;
  }

  get isMultiScale() {
    return typeof this.args.xScale === 'object';
  }

  get hasLineText() {
    return this.hoverTextValues && this.hoverTextValues.length > 0;
  }

  get cornerTexts() {
    if (!this.cornerTextValues) return []
    return this.cornerTextValues.map((value) => {
      if(value instanceof Date) {
        return this.formatDate(value);
      } else if(typeof value === 'undefined') {
        return ""
      } else {
        return `${value}${this.cornerTextUnits}`;
      }
    });
  }

  // ~~~~~~~~~~~~~~~~~~~~
  // INTERACTION EVENTS
  // ~~~~~~~~~~~~~~~~~~~~

  mousemove = (e) => {
    throttle(this, this.updateLine, e, 20);
  }

  mouseout = () => {
    this.fadeState = "opacity-0";
  }

  mouseover = (e) => {
    if(e.type === "touchstart") {
      this.currentMousePos = this.mouseOrTouchPosition(e);
      this.setMovingElementsPosition(this.currentMousePos);
    }
    this.fadeState = "opacity-100";
  }

  initElement = (element, [elementName]) => {
    this[elementName] = element;
  }

  // Move line and update text values
  updateLine = (e) => {
    const mousePos = this.mouseOrTouchPosition(e);
    this.currentMousePos = mousePos;
    if( mousePos > 0.5 ) {
      const nearestPoints = this.findNearestDatapoints(mousePos);
      if ( !this.putValueInCorner ) this.updateHoverText(nearestPoints);
      this.updateCornerText(nearestPoints);
    }
  }

  updateHoverText(nearestPoints) {
    const yLoc = this.findMin(nearestPoints, 'y');
    this.textBox = this.findTextLocAndSize(nearestPoints, yLoc);
    this.hoverTextValues = nearestPoints.map((datapoint, index) => {
      return {
        ...datapoint,
        y: this.textBox.y + this.textBoxPadding + (index * this.textLineHeight + 10)
      };
    });
  }

  // update corner text elements
  updateCornerText(nearestPoints) {
    const dateValues =  this.singleValueCornerText
              ? [ typeof nearestPoints[0][this.dateKey] === 'undefined' ? new Date(nearestPoints[0][this.dateKey]) : nearestPoints[0][this.dateKey] ]
      : nearestPoints.map((d) => typeof d[this.dateKey] === "undefined" ? d[this.dateKey] : new Date(d[this.dateKey]));
    const valueTextElements = this.putValueInCorner ? nearestPoints.map((point) => {
      return point.text
    }) : []
    this.cornerTextValues = [...dateValues, ...valueTextElements]
  }

  // For each dataset, return nearest dataset points to scrubber, alongside x position, y position, color, and text value
  findNearestDatapoints(mousePos) {
    if (!this.args.datasets) return [{x: "", y: ""}];

    return this.args.datasets.map((dataset, index) => {
      const xScale = this.isMultiScale ? this.args.xScale[index] : this.args.xScale;
      const x0 = xScale.invert(mousePos); // x position within the element.
      const dataPoint = this.nearestDataPoint(dataset, x0);
      const time = dataPoint
        ? typeof dataPoint[this.dateKey] === "string" ? new Date(dataPoint[this.dateKey]) : dataPoint[this.dateKey]
        : 0 
      const x = dataPoint ? xScale(time) : 0;
      return {
        ...dataPoint,
        x: x < 0 ? 0 : x,
        y: dataPoint ? this.args.yScale(dataPoint[this.valTextKey]) - this.textLineHeight : 0,
        color: this.args.colors ? this.args.colors[index] : 'black',
        text: this.calculateText(dataPoint ? dataPoint[this.valTextKey] : null)
      }
    });
  }

  // Get nearest dataset point to date x0
  nearestDataPoint(dataset, x0) {
    const i = this.bisectDate(dataset, x0, 1);
    const d0 = dataset[i - 1];
    const d1 = dataset[i];
    // out of dataset bounds
    if(dataset.length === 0 || x0 < dataset[0].at || x0 > dataset[dataset.length - 1].at) return
    return (d1 && d0) ? (x0 - d0.at > d1.at - x0 ? d1 : d0) : 0;
  }

  // Find textbox y position and dimensions
  findTextLocAndSize(dataPoints, yLoc) {
    if(!dataPoints.length) return 0
    const pointWithLargestText = dataPoints.reduce((max, dataPoint) => !max || dataPoint.text.length > max ? dataPoint : max);
    const textSize = this.calculateTextDimensions(pointWithLargestText.text);
    const textBoxWidth = this.textBoxPadding * 2 + textSize.width
    const y = yLoc - (this.textBoxPadding * 4 + (this.textLineHeight * dataPoints.length))
    return {
      y: y < 5 ? 5 : y,
      width: textBoxWidth,
      height: this.textLineHeight * dataPoints.length + this.textBoxPadding * 2,
      translate: textBoxWidth / 2
    }
  }  

  // Returns a mouse or touch position relative to the svg
  mouseOrTouchPosition(event) {
    const clientX = event.type.includes("touch") ? event.targetTouches[0].clientX : event.clientX;
    return clientX - this.findSvg().getBoundingClientRect().left;
  }

  // Request animation frame loop that serves to smooth and animate line to a target mouse position
  animateScrubbing = () => {
    const x = (this.currentMousePos || this.currentMousePos === 0)
            ? this.getAnimatedLocation(this.line, 'x1')
            : 0
    if(x) {
      this.setMovingElementsPosition(x);
    }
  }

  // If there is a mouse position, this returns an x position based on mouse position and dx
  getAnimatedLocation(element, attrName) {
    const currentX = parseInt(element.getAttribute(attrName)); // Get current position
    const totalDx = this.currentMousePos - currentX; // Find total x distance from cursor to element
    const acceleration = 4; // acceleration scalar
    const frameDx = totalDx / acceleration; // delta x for this frame

    return Math.abs(frameDx) > 1                      // if the delta x is less than 1 it needs to be
            ? currentX + frameDx                      // rounded or else it will register as 0px and get stuck
            : totalDx < 1 ? this.currentMousePos : currentX + 1; // note only happens really close to cursor as it deccelerates
  }

  // Move line to position x
  setMovingElementsPosition = (x) => {
    this.line.setAttribute('x1', x);
    this.line.setAttribute('x2', x);
    if (!this.putValueInCorner) this.moveAndBoundTextBox(x);
  }

  // Move scrubber text box but bound it by the edges of the svg chart
  moveAndBoundTextBox(x) {
    const left = x - this.textBox.width/2;
    const right = x + this.textBox.width / 2;
    let boundedBoxPosition;

    // If past the left keep box inside chart
    if (left < 5) {
      boundedBoxPosition = 5;
    // If past the right side of the chart, keep inside the chart
    } else if (right > (this.chartWidth - 5)) {
      boundedBoxPosition = left - (right - this.chartWidth);
    } else {
      boundedBoxPosition = left;
    }

    this.movingTextContainer.setAttribute('x', boundedBoxPosition);
    this.animatedTextXPosition = boundedBoxPosition + this.textBox.width/2;
  }

  // Find text values based on value type and unit configuration
  calculateText(value) {
    if( value == null ) return "";
    if(value instanceof Date) {
      return this.formatDate(value);
    } else {
      return `${parseFloat(value).toFixed(2)}${this.lineTextUnits}`; 
    }
  }

  findMin(array, field) {
    return array.reduce((max, obj) => {
      return obj && obj[field] < max ? obj[field] : max;
    }, this.args.yScale(0));
  }

  formatDate(date) {
    const dateObj = new Date(date);
    const hours = dateObj.getHours();
    const minutes = dateObj.getMinutes();
    const foramttedDate = this.hideDate ? '' : `${dateObj.getMonth() + 1}-${dateObj.getDate()}-${dateObj.getFullYear()}`;
    const time = `${this.hideDate ? '' : ', '}${hours < 12 ? hours : hours - 12}:${('0' + minutes).slice(-2)} ${hours < 12 ? "am" : "pm"}`;
    return `${foramttedDate}${this.hideTime ? "" : time}`;
  }

  calculateTextDimensions(text) {
    this.textSizeElement.innerHTML = text;
    const size = this.textSizeElement.getBoundingClientRect();
    return {
      width: size.width + 1,
      height: size.height + 1
    }
  }

  // Get parent svg element based on inited element
  findSvg() {
    return this.line.closest('svg')
  }
}